# coding: utf-8

from __future__ import print_function

import licant
from licant.core import UpdatableTarget, UpdateStatus, core
from licant.cache import fcache
from licant.util import purple, quite
import threading
import os

_rlock = threading.RLock()


def do_execute(target, rule, msgfield, prefix=None):
    def sprint(*args, **kwargs):
        """print for multithread build"""
        with _rlock:
            print(*args, **kwargs)

    rule = rule.format(**target.__dict__)

    message = getattr(target, msgfield, None)

    if not core.runtime["quite"]:
        if not core.runtime["debug"] and message is not None:
            if not isinstance(message, quite):
                if prefix is not None:
                    sprint(prefix, message.format(**target.__dict__))
                else:
                    sprint(message.format(**target.__dict__))

        else:
            sprint(rule)

    ret = os.system(rule)

    if target.isfile:
        target.update_info(target)

    return True if ret == 0 else False


class Executor:
    def __init__(self, rule, msgfield="message"):
        self.rule = rule
        self.msgfield = msgfield

    def __call__(self, target, **kwargs):
        return do_execute(target, self.rule, self.msgfield, **kwargs)


class MakeFileTarget(UpdatableTarget):
    __actions__ = UpdatableTarget.__actions__.union(
        {"actlist", "makefile", "clean"})

    def __init__(self, tgt, deps, **kwargs):
        UpdatableTarget.__init__(self, tgt, deps, **kwargs)
        self.default_action = "makefile"

    def clean(self):
        stree = self.core.subtree(self.tgt)
        return stree.invoke_foreach(ops="clr", cond=if_file_and_exist)

    def makefile(self):
        return self.recurse_update()


class FileTarget(MakeFileTarget):
    __actions__ = MakeFileTarget.__actions__.union({"build", "clr"})

    def __init__(self, tgt, deps, force=False, **kwargs):
        MakeFileTarget.__init__(self, tgt, deps, **kwargs)
        self.isfile = True
        self.force = force
        self.clrmsg = "DELETE {tgt}"

    def update_info(self, _self):
        fcache.update_info(self.tgt)
        return True

    def mtime(self):
        curinfo = fcache.get_info(self.tgt)
        if not curinfo.exist:
            return 0
        else:
            return curinfo.mtime

    def dirkeep(self):
        """Create directory tree for this file if needed."""
        dr = os.path.normpath(os.path.dirname(self.tgt))
        if not os.path.exists(dr):
            print("MKDIR %s" % dr)
            os.makedirs(dr)
        return True

    def is_exist(self):
        curinfo = fcache.get_info(self.tgt)
        return curinfo.exist

    def warn_if_not_exist(self):
        """Print warn if file isn't exist"""
        info = fcache.get_info(self.tgt)
        if not info.exist:
            print("Warn: file {} isn`t exist".format(purple(self.tgt)))

    def clr(self):
        """Delete this file."""
        do_execute(self, "rm -f {tgt}", "clrmsg")

    def self_need(self):
        if self.force or not self.is_exist():
            return True

        maxtime = 0
        for dep in self.get_deplist():
            if dep.mtime() > maxtime:
                maxtime = dep.mtime()

        if maxtime > self.mtime():
            return True

        return False

    def update(self):
        self.dirkeep()
        return self.build(self)


class FileSet(MakeFileTarget):
    """Virtual file target`s set.

        For link a set of file objects to the licant tree
        without depend`s overhead."""

    def __init__(self, tgt, targets, deps, **kwargs):
        MakeFileTarget.__init__(self, tgt=tgt, deps=targets+deps, **kwargs)
        self.targets = targets
        self.__mtime = None

    def self_need(self):
        return False

    def update(self):
        pass

    def mtime(self):
        if self.__mtime is None:
            maxtime = 0
            for dep in self.get_deplist():
                if dep.mtime() > maxtime:
                    maxtime = dep.mtime()
            self.__mtime = maxtime

        return self.__mtime


def source(tgt, deps=[]):
    """Index source file by licant core."""

    if core.exist(tgt):
        return

    target = FileTarget(
        build=lambda self: self.warn_if_not_exist(), deps=deps, tgt=tgt)
    target.clr = None
    target.dirkeep = licant.util.do_nothing
    target.update_status = UpdateStatus.Keeped
    core.add(target)
    return tgt


def copy(tgt, src, adddeps=[], message="COPY {src} {tgt}"):
    """Make the file copy target."""
    src = os.path.expanduser(src)
    tgt = os.path.expanduser(tgt)

    source(src)

    core.add(
        FileTarget(
            tgt=tgt,
            build=Executor("cp {src} {tgt}"),
            src=src,
            deps=[src] + adddeps,
            message=message,
        )
    )
    return tgt


def makefile(tgt, deps, do, **kwargs):
    """Make the file copy target."""
    core.add(
        FileTarget(
            tgt=tgt,
            build=Executor(do),
            deps=deps,
            **kwargs
        )
    )


def fileset(tgt, targets, deps=[], **kwargs):
    """Make a fileset."""
    core.add(FileSet(tgt=tgt, targets=targets, deps=deps, **kwargs))
    return tgt


def if_file_and_exist(target):
    if not isinstance(target, FileTarget):
        return False

    curinfo = fcache.get_info(target.tgt)
    return curinfo.exist



from licant.core import core
import licant.make
import os


class java_options:
    def __init__(
            self,
            toolchain=None,
            javac_flags=""
    ):
        self.toolchain = toolchain
        self.javac_flags = javac_flags
        self.javacrule = "{opts.toolchain.javac} {opts.javac_flags} {srcs}"
        self.dexrule = "{opts.toolchain.dx} --dex --output {tgt} {src}"


class java_toolchain:
    def __init__(self, javac):
        self.javac = javac


class android_toolchain:
    def __init__(self, javac, dx):
        self.javac = javac
        self.dx = dx


def javac(srcs, tgt, opts=java_options(), message="JAVA {tgt}", deps=[]):
    build = licant.make.Executor(opts.javacrule)

    for j in srcs:
        licant.make.source(j)

    core.add(
        licant.make.FileTarget(
            opts=opts,
            tgt=tgt,
            srcs=" ".join(srcs),
            deps=deps + srcs,
            build=build,
            message=message
        )
    )


def dex(src, tgt, opts=java_options(), message="DEX {tgt}", deps=[]):
    build = licant.make.Executor(opts.dexrule)

    # for j in srcs:
    #	licant.make.source(j) // scan class files

    core.add(
        licant.make.FileTarget(
            opts=opts,
            tgt=tgt,
            src=src,
            deps=deps,
            build=build,
            message=message
        )
    )



from licant.util import red
from licant.scripter import scriptq
import licant.core

import re

special = ["__script__", "__dir__"]


class Module:
    def __init__(self, name, script=None, dir=None, stack=None, **kwargs):
        self.name = name
        self.script = script
        self.stack = stack
        self.opts = kwargs
        self.opts["__script__"] = self.script
        self.opts["__dir__"] = dir


class VariantModule:
    def __init__(self):
        self.impls = {}

    def addimpl(self, impl, mod):
        self.impls[impl] = mod
        mod.impl = impl


class ModuleLibrary:
    def __init__(self):
        self.modules = {}
        self.defimpls = {}

    def register_impl(self, mod, impl):
        if mod.name in self.modules:
            if not isinstance(self.modules[mod.name], VariantModule):
                print("Attempt to register the module {} again".format(red(mod.name)))
                exit(-1)
            else:
                varmod = self.modules[mod.name]
        else:
            varmod = VariantModule()
            self.modules[mod.name] = varmod

        varmod.addimpl(impl, mod)

    def get(self, name, impl=None):
        if name not in self.modules:
            print("The missing module {} was requested".format(red(name)))
            exit(-1)

        m = self.modules[name]
        if impl is None:
            return self.get_default(name)
        else:
            if isinstance(m, Module):
                licant.error("stand by module is deprecated")
            else:
                if impl in m.impls:
                    return m.impls[impl]
                else:
                    if impl == "__none__":
                        m.addimpl(impl="__none__", mod=Module(name))
                        return m.impls[impl]

                    print("Unregistred implementation: {} (module:{})".format(
                        red(impl), red(name)))
                    exit(-1)

    def set_defimpl(self, modname, impl, force=False):
        if modname in self.defimpls and not force:
            licant.error(
                "Default implementation for module {} setted twice : first:{}, second:{}".format(
                    licant.util.yellow(modname),
                    licant.util.yellow(
                        self.defimpls[modname] + ":" + self.get(modname, self.defimpls[modname]).stack[-1]),
                    licant.util.yellow(
                        impl + ":" + self.get(modname, impl).stack[-1]),
                )
            )
        self.defimpls[modname] = impl

    def is_variant(self, name):
        licant.error("deprecated with 'always implementation concept'")
        if name not in self.modules:
            licant.error(
                "Isn't registred as module ({})".format(
                    licant.util.yellow(name))
            )
        return isinstance(self.modules[name], VariantModule)

    def get_default(self, name):
        if name not in self.defimpls:
            if name not in self.modules:
                licant.error(
                    "Unregistred module ({})".format(
                        licant.util.yellow(name)
                    )
                )

            else:
                licant.error(
                    "Doesn`t have default impl of that module ({})".format(
                        licant.util.yellow(name)
                    )
                )

        return self.get(name, self.defimpls[name])


mlibrary = ModuleLibrary()


def module(name, impl=None, default=False, **kwargs):
    if impl is not None:
        implementation(name, impl, default=default, **kwargs)
        return
    implementation(name, "__default__", default=True, **kwargs)


def implementation(name, impl, default=False, **kwargs):
    mlibrary.register_impl(
        Module(
            name,
            script=scriptq.last(),
            dir=scriptq.curdir(),
            stack=list(scriptq.stack),
            **kwargs
        ),
        impl=impl,
    )

    if default:
        module_default_implementation(name, impl)


class submodule:
    def __init__(self, name, impl=None, addopts=None):
        if isinstance(name, tuple):
            self.name = name[0]
            self.impl = name[1]
            self.addopts = addopts
        elif isinstance(name, submodule):
            self.name = name.name
            self.impl = name.impl
            self.addopts = addopts
        else:
            self.name = name
            self.impl = impl
            self.addopts = addopts

    def __repr__(self):
        return "subm(" + self.name + ")"


def module_default_implementation(mod, impl):
    mlibrary.set_defimpl(mod, impl, force=True)


def print_modules_list(target, *args):
    if len(mlibrary.modules) == 0:
        print("modules doesn't founded")
        return

    mkeys = sorted(mlibrary.modules.keys())

    if len(args) > 0:
        mkeys = [m for m in mkeys if re.search(args[0], m)]

    for k in mkeys:
        v = mlibrary.modules[k]
        if isinstance(v, VariantModule):
            print("{}: {}".format(k, list(v.impls.keys())))


def print_module(target, *args):
    if len(args) == 0:
        print("Usage: EXECUTABLE m print NAME\nUsage: EXECUTABLE m print NAME IMPL\n")
        return

    if len(args) == 1:
        module = mlibrary.get_default(args[0])
    else:
        module = mlibrary.get(args[0], args[1])

    print(module.__dict__)


modules_target = licant.core.Target(
    tgt="m",
    deps=[],
    list=print_modules_list,
    print=print_module,
    actions={"list", "print"},
    __help__="Info about modules",
)

licant.core.core.add(modules_target)



from licant.scripter import scriptq
from licant.util import yellow
import licant.core

import os
import json

gpath = "/var/lib/licant"
lpath = os.path.expanduser("~/.licant")

libs = None

included = dict()


def merge_two_dicts(x, y):
    z = x.copy()  # start with x's keys and values
    z.update(y)  # modifies z with y's keys and values & returns None
    return z


def init():
    global libs

    glibs = {}
    llibs = {}

    if os.path.exists(gpath):
        glibs = json.load(open(gpath))

    if os.path.exists(lpath):
        llibs = json.load(open(lpath))

    libs = merge_two_dicts(glibs, llibs)


def include(lib, path=None, local_tunel=None):
    if libs is None:
        init()

    if lib in included:
        if path and path != included[lib]:
            print(
                "Warning: prevent library by path({}) include becouse library been include early with another path({})".format(path, included[lib]))
        return

    if path is not None:
        included[lib] = path
        print("LICANTLIB", lib, path)
        scriptq.execute(path)
        return

    if local_tunel is not None:
        # Local tunel is special technique for use licant
        # with python packages. Local tunel create link
        # for setup.py can find sources by relative path.

        if not os.path.exists(local_tunel[0]):
            # Убеждаемся, что директории до тунеля существуют
            if not os.path.exists(os.path.dirname(local_tunel[0])):
                os.makedirs(os.path.dirname(local_tunel[0]))

            rawdir = os.path.dirname(libs[lib])

            if not os.path.exists(local_tunel[0]):
                os.symlink(rawdir, local_tunel[0])

            included[lib] = os.path.join(local_tunel[0], local_tunel[1])
            scriptq.execute(os.path.join(local_tunel[0], local_tunel[1]))
            return

        else:
            included[lib] = os.path.join(local_tunel[0], local_tunel[1])
            scriptq.execute(os.path.join(local_tunel[0], local_tunel[1]))
            return

    if lib not in libs:
        print(
            "Unregistred library {}. Use licant-config utility or manually edit {} or {} file. NOTE: If you use local libraries, maybe you need to reorder your includes".format(
                yellow(lib), yellow(lpath), yellow(gpath)
            )
        )
        exit(-1)

    included[lib] = libs[lib]
    print("LICANTLIB", lib, included[lib])
    scriptq.execute(libs[lib])


def print_system_libs(taget, *args):
    if libs is None:
        init()

    keys = sorted(libs.keys())
    for k in keys:
        print("{}: {}".format(k, libs[k]))


def print_included_libs(taget, *args):
    keys = sorted(included.keys())
    for k in keys:
        print("{}: {}".format(k, included[k]))


libs_target = licant.core.Target(
    tgt="l",
    deps=[],
    list=print_system_libs,
    included=print_included_libs,
    actions={
        "list",
        "included"
    }, __help__="Licant libs info"
)

licant.core.core.add(libs_target)



from licant.core import core
from licant.util import yellow
import licant.make
import os


class Options:
    pass


class toolchain:
    def __init__(
            self, cxx, cc, ld, ar,
            objdump, moc=None, uic=None, objcopy="objcopy"):
        self.cc = cc
        self.cxx = cxx
        self.ld = ld
        self.ar = ar
        self.objdump = objdump
        self.objcopy = objcopy
        self.moc = moc
        self.uic = uic

    def __repr__(self):
        return str(self.__dict__)


host_toolchain = toolchain(
    cxx="c++",
    cc="cc",
    ld="ld",
    ar="ar",
    objdump="objdump",
    moc="moc",
    uic="uic",
    objcopy="objcopy"
)


def toolchain_gcc(prefix):
    print("{} is deprecated. Use {} instead".format(
        yellow('toolchain_gcc'),
        yellow('gcc_toolchain')
    ))
    return gcc_toolchain(prefix)


def gcc_toolchain(prefix):
    if prefix is None:
        return host_toolchain

    return toolchain(
        cc=prefix+"gcc",
        cxx=prefix+"g++",
        ld=prefix+"ld",
        ar=prefix+"ar",
        objdump=prefix+"objdump",
        objcopy=prefix+"objcopy",
        moc="moc",
        uic="uic")


def clang_toolchain():
    return toolchain(
        cc="clang++",
        cxx="clang",
        ld="ld",
        ar="ar",
        objdump="objdump",
        objcopy="objcopy",
        moc="moc",
        uic="uic")


class options:
    def __init__(
            self,
            toolchain=host_toolchain,
            include_paths=None,
            defines=None,
            cxx_flags="",
            cc_flags="",
            ld_flags="",
            ld_srcs_add="",
            ldscripts=None,
            optimize="",
    ):
        self.toolchain = toolchain
        self.incopt = licant.util.flag_prefix("-I", include_paths)
        self.defopt = licant.util.flag_prefix("-D", defines)
        self.ldscripts = licant.util.flag_prefix("-T", ldscripts)
        self.cxx_flags = cxx_flags
        self.cc_flags = cc_flags
        self.ld_flags = ld_flags
        self.ld_srcs_add = ld_srcs_add
        self.optimize = optimize
        self.execrule = "{opts.toolchain.cxx} {opts.ld_flags} -Wl,--start-group {srcs} {opts.ld_srcs_add} -Wl,--end-group -o {tgt} {opts.ldscripts}"
        self.dynlibrule = "{opts.toolchain.cxx} --shared {opts.ld_flags} -Wl,--start-group {srcs} {opts.ld_srcs_add} -Wl,--end-group -o {tgt} {opts.ldscripts}"
        self.statlibrule = "{opts.toolchain.ar} rcs {tgt} {srcs}"
        self.cxxobjrule = "{opts.toolchain.cxx} -c {src} -o {tgt} {opts.incopt} {opts.defopt} {opts.cxx_flags} {opts.optimize}"
        self.ccobjrule = "{opts.toolchain.cc} -c {src} -o {tgt} {opts.incopt} {opts.defopt} {opts.cc_flags} {opts.optimize}"
        self.cxxdeprule = "{opts.toolchain.cxx} -MM {src} > {tgt} {opts.incopt} {opts.defopt} {opts.cxx_flags}"
        self.ccdeprule = "{opts.toolchain.cc} -MM {src} > {tgt} {opts.incopt} {opts.defopt} {opts.cc_flags}"
        self.mocrule = "{opts.toolchain.moc} {src} > {tgt}"
        self.uicrule = "{opts.toolchain.uic} {src} > {tgt}"

    def __str__(self):
        return "(toolchain:{}, incopt:{}, defopt:{}, cxx_flags:{}, cc_flags:{}, ld_flags:{}, ld_srcs_add:{}, ldscripts:{})".format(
            self.toolchain, self.incopt, self.defopt, self.cxx_flags, self.cc_flags, self.ld_flags, self.ld_srcs_add, self.ldscripts
        )


cxx_ext_list = ["cpp", "cxx"]
cc_ext_list = ["cc", "c"]
asm_ext_list = ["asm", "s", "S"]


def object(src, tgt, opts=options(), type=None, deps=None, message="OBJECT {tgt}"):
    if deps is None:
        deps = [src]

    if type is None:
        ext = os.path.basename(src).split(".")[-1]

        if ext in cxx_ext_list:
            type = "cxx"
        elif ext in cc_ext_list:
            type = "cc"
        elif ext in asm_ext_list:
            type = "asm"
        else:
            print("Unrecognized extention: {}".format(licant.util.red(ext)))
            exit(-1)
    if type == "cxx":
        build = licant.make.Executor(opts.cxxobjrule)
    elif type == "cc":
        build = licant.make.Executor(opts.ccobjrule)
    elif type == "asm":
        build = licant.make.Executor(opts.ccobjrule)
    else:
        print(licant.util.red("Unrecognized extention"))
        exit(-1)
    core.add(
        licant.make.FileTarget(
            opts=opts, tgt=tgt, src=src, deps=deps, build=build, message=message
        )
    )


def qt_moc(src, tgt, opts=options(), type=None, deps=None, message="MOC {tgt}"):
    if deps is None:
        deps = [src]

    core.add(
        licant.make.FileTarget(
            opts=opts,
            tgt=tgt,
            src=src,
            deps=deps,
            build=licant.make.Executor(opts.mocrule),
            message=message,
        )
    )


def qt_uic(src, tgt, opts=options(), type=None, deps=None, message="UIC {tgt}"):
    if deps is None:
        deps = [src]

    core.add(
        licant.make.FileTarget(
            opts=opts,
            tgt=tgt,
            src=src,
            deps=deps,
            build=licant.make.Executor(opts.uicrule),
            message=message,
        )
    )


def depend(
        src, tgt, opts=options(), type=None, deps=None, message="DEPENDS {tgt}", **kwargs
):
    if deps is None:
        deps = [src]

    if type is None:
        ext = os.path.basename(src).split(".")[-1]

        if ext in cxx_ext_list:
            type = "cxx"
        elif ext in cc_ext_list:
            type = "cc"
        elif ext in asm_ext_list:
            type = "asm"
        else:
            print("Unrecognized extention: {}".format(licant.util.red(ext)))
            exit(-1)
    if type == "cxx":
        build = licant.make.Executor(opts.cxxdeprule)
    elif type == "cc":
        build = licant.make.Executor(opts.ccdeprule)
    elif type == "asm":
        build = licant.make.Executor(opts.ccdeprule)
    else:
        print(licant.util.red("Unrecognized extention"))
        exit(-1)

    core.add(
        licant.make.FileTarget(
            opts=opts,
            tgt=tgt,
            src=src,
            deps=deps,
            build=build,
            message=message,
            **kwargs
        )
    )


def executable(tgt, srcs, opts=options(), message="EXECUTABLE {tgt}"):
    core.add(
        licant.make.FileTarget(
            opts=opts,
            tgt=tgt,
            build=licant.make.Executor(opts.execrule),
            srcs=" ".join(srcs),
            deps=srcs,
            message=message,
        )
    )


def dynamic_library(tgt, srcs, opts=options(), message="DYNLIB {tgt}"):
    core.add(
        licant.make.FileTarget(
            opts=opts,
            tgt=tgt,
            build=licant.make.Executor(opts.dynlibrule),
            srcs=" ".join(srcs),
            deps=srcs,
            message=message,
        )
    )


def static_library(tgt, srcs, opts=options(), message="STATLIB {tgt}"):
    core.add(
        licant.make.FileTarget(
            opts=opts,
            tgt=tgt,
            build=licant.make.Executor(opts.statlibrule),
            srcs=" ".join(srcs),
            deps=srcs,
            message=message,
        )
    )


def make_gcc_binutils(pref):
    return toolchain(
        cxx=pref + "-g++",
        cc=pref + "-gcc",
        ld=pref + "-ld",
        ar=pref + "-ar",
        objdump=pref + "-objdump",
    )


def disassembler(target, *args):
    if len(args) <= 1:
        print("usage: disasm object_path asmlist_path")

    _target = core.get(args[0])
    _target.makefile()

    cmd = "{} -D {} > {}".format(_target.opts.binutils.objdump,
                                 _target.tgt, args[1])
    print(cmd)
    os.system(cmd)


def objcopy(toolchain, tgt, src, format, sections, message="OBJCOPY {tgt}"):
    sections_str = " ".join(["-j {}".format(s) for s in sections])
    rule = "{opts.toolchain.objcopy} -O {opts.format} {opts.sections_str} {opts.src} {opts.tgt}"

    opts = Options()
    opts.toolchain = toolchain
    opts.sections_str = sections_str
    opts.format = format
    opts.src = src
    opts.tgt = tgt

    core.add(
        licant.make.FileTarget(
            opts=opts,
            deps=[src],
            tgt=tgt,
            build=licant.make.Executor(rule),
            message=message,
        )
    )


binutils_target = licant.core.Target(
    tgt="binutils", deps=[], disasm=disassembler, actions={"disasm"}
)

core.add(binutils_target)



import os


class FileCache:
    def __init__(self):
        self.cache = dict()

    class fileinfo:
        def __init__(self, path):
            self.exist = os.path.exists(path)
            if self.exist:
                self.mtime = os.stat(path).st_mtime
            else:
                self.mtime = None

    def update_info(self, path):
        self.cache[path] = self.fileinfo(path)

    def get_info(self, path):
        if path not in self.cache:
            self.update_info(path)
        return self.cache[path]


fcache = FileCache()



from licant.cli import cliexecute as ex

from licant.core import Core, Target, UpdatableTarget, UpdateStatus
from licant.core import core as default_core
from licant.core import routine_decorator as routine

from licant.core import do
from licant.core import get_target

from licant.make import copy, fileset, makefile, source
from licant.cxx_make import objcopy
from licant.cxx_modules import application as cxx_application
from licant.cxx_modules import shared_library as cxx_shared_library
from licant.cxx_modules import static_library as cxx_static_library
from licant.cxx_modules import library as cxx_library
from licant.cxx_modules import objects as cxx_objects

from licant.modules import module, implementation, submodule
from licant.modules import module_default_implementation as module_defimpl
from licant.modules import module_default_implementation
from licant.util import error

from licant.cxx_make import gcc_toolchain, clang_toolchain, host_toolchain

import licant.scripter
from licant.libs import include

__version__ = "1.9.4"


def directory():
    return licant.scripter.scriptq.curdir()


def subtree(tgt):
    return default_core.subtree(tgt)


def execute(path):
    licant.scripter.scriptq.execute(path)


def execute_recursive(*argv, **kwargs):
    licant.scripter.scriptq.execute_recursive(*argv, **kwargs)


def about():
    return "I'm Licant"


class Object(object):
    pass


glbfunc = Object()
attribute_store = Object()


def global_function(var):
    setattr(glbfunc, var.__name__, var)


def import_attribute(name, var):
    setattr(attribute_store, name, var)


def attribute(name):
    return getattr(attribute_store, name)


__all__ = [
    "include",
    "error",
    "module_default_implementation",
    "module_defimpl",
    "module",
    "implementation",
    "submodule",

    "cxx_modules",
    "cxx_objects",
    "cxx_library",
    "cxx_application",
    "shared_library",
    "static_library",

    "makefile",
    "fileset",
    "copy",
    "source",
    "do",
    "routine",
    "routine_decorator",

    "UpdateStatus",
    "UpdatableTarget",
    "Core",
    "core",
    "cxx_static_library",
    "cxx_shared_library",
    "objcopy",
    "get_target",
    "Target",
    "ex"
]



#!/usr/bin/env python3

from setuptools import setup
import glob
import os

import licant

setup(
    name="licant",
    packages=["licant"],
    version=licant.__version__,
    license="MIT",
    description="licant make system",
    author="Sorokin Nikolay",
    author_email="mirmikns@yandex.ru",
    url="https://github.com/mirmik/licant",
    long_description=open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    keywords=["testing", "make"],
    classifiers=[],
    scripts=["configurator/licant-libs"],
    # 	package_data={'licant': [
    #   	'templates/cxx/make.py',
    # 	  	'templates/cxx/main.cpp',
    # 	 	'templates/cxxgxx/make.py',
    # 		'templates/cxxgxx/main.cpp',
    #    ]}
)



# coding: utf-8

import licant.util
import threading
import types
import inspect
from enum import Enum
import re


class WrongAction(Exception):
    def __init__(self, obj, actname):
        self.obj = obj
        self.actname = actname

    def __str__(self):
        return "WrongAction: obj:{} actname:{} class:{} dict:{self.obj.__dict__}".format(self.obj, self.actname, self.obj.__class__, self.obj.__dict__)


class NoneDictionary(dict):
    def __init__(self):
        dict.__init__(self)

    def __getitem__(self, idx):
        try:
            return dict.__getitem__(self, idx)
        except Exception:
            return None


class Core:
    def __init__(self):
        self.targets = {}
        self.help_showed_targets = []
        self.runtime = NoneDictionary()

    def exist(self, name):
        return name in self.targets

    def add(self, target):
        """Add new target"""
        target.core = self
        self.targets[target.tgt] = target

        if target.__help__ is not None:
            self.help_showed_targets.append(target)

    def get(self, tgt):
        """Get target object"""
        if tgt in self.targets:
            return self.targets[tgt]
        licant.util.error("unregistred target " + licant.util.yellow(tgt))

    def subtree(self, root):
        """Construct Subtree accessor for root target"""
        return SubTree(self, root)

    def depends_as_set(self, tgt, incroot=True):
        """TODO: as_set, but list returned???"""
        res = set()
        if incroot:
            res.add(tgt)

        target = self.get(tgt)

        for d in target.deps:
            if d not in res:
                res.add(d)
                subres = self.depends_as_set(d)
                res = res.union(subres)
        return sorted(res)


# Объект ядра с которым библиотеки работают по умолчанию.
core = Core()


class SubTree:
    def __init__(self, core, root):
        self.root = root
        self.core = core
        self.depset = core.depends_as_set(root)
        self.weakdepsset = list(get_target(root).weakdeps)

    #    def update(self):
    #        SubTree.__init__(self, root)

    def invoke_foreach(self, ops, cond=licant.util.always_true):
        if core.runtime["debug"]:
            print("FOREACH(root={}, ops={}, cond={})".format(self.root, ops, cond))

        sum = 0
        ret = None

        for d in self.depset + self.weakdepsset:
            target = self.core.get(d)
            if cond(target):
                ret = target.invoke(ops)

            if ret is not None:
                sum += 1

        return sum

    def __generate_rdepends_lists(self, targets):
        for t in targets:
            t.rdepends = []
            t.rcounter = 0

        for t in targets:
            for dname in t.deps:
                dtarget = self.core.get(dname)
                dtarget.rdepends.append(t.tgt)

    def reverse_recurse_invoke_single(
        self, ops, threads=None, cond=licant.util.always_true
    ):
        if core.runtime["trace"]:
            print(
                "SINGLETHREAD MODE(root={}, ops={}, cond={})".format(
                    self.root, ops, cond
                )
            )
        targets = [self.core.get(t) for t in self.depset]

        self.__generate_rdepends_lists(targets)

        works = licant.util.queue()

        for t in targets:
            if t.rcounter == len(t.deps):
                works.put(t)

        while not works.empty():
            w = works.get()

            if cond(w):
                ret = w.invoke(ops)
                if ret is False:
                    print(licant.util.red("runtime error"))
                    exit(-1)

            for r in [self.core.get(t) for t in w.rdepends]:
                r.rcounter = r.rcounter + 1
                if r.rcounter == len(r.deps):
                    works.put(r)

    def reverse_recurse_invoke_threads(
        self, ops, threads, cond=licant.util.always_true
    ):
        if core.runtime["trace"]:
            print(
                "MULTITHREAD MODE(root={}, threads={}, ops={}, cond={})".format(
                    self.root, threads, ops, cond
                )
            )

        targets = [self.core.get(t) for t in self.depset]

        self.__generate_rdepends_lists(targets)
        works = licant.util.queue()

        class info_cls:
            def __init__(self):
                self.have_done = 0
                self.need_done = len(targets)
                self.sum = 0
                self.err = False

        info = info_cls()

        for t in targets:
            if t.rcounter == len(t.deps):
                works.put(t)

        lock = threading.Lock()

        def thread_func(index):
            while info.have_done != info.need_done:
                if info.err:
                    break

                lock.acquire()
                if not works.empty():
                    w = works.get()
                    lock.release()

                    if core.runtime["trace"]:
                        print(
                            "TRACE: THREAD {0} get work {1}".format(index, w))

                    if cond(w):
                        try:
                            ret = w.invoke(ops)
                        except Exception:
                            info.err = True
                            return
                        if ret is False:
                            info.err = True
                            return
                        if ret == 0:
                            info.sum += 1

                    for r in [self.core.get(t) for t in w.rdepends]:
                        r.rcounter = r.rcounter + 1
                        if r.rcounter == len(r.deps):
                            works.put(r)

                    info.have_done += 1

                    if core.runtime["trace"]:
                        print(
                            "TRACE: THREAD {0} finished with work {1}".format(
                                index, w)
                        )

                    continue
                lock.release()

        threads_list = [
            threading.Thread(target=thread_func, args=(i,)) for i in range(0, threads)
        ]

        for t in threads_list:
            t.start()

        for t in threads_list:
            t.join()

        if info.err:
            print(licant.util.red("runtime error (multithreads mode)"))
            exit(-1)
        return info.sum

    def reverse_recurse_invoke(self, *args, **kwargs):
        if "threads" in kwargs:
            if kwargs["threads"] == 1:
                return self.reverse_recurse_invoke_single(*args, **kwargs)
            else:
                return self.reverse_recurse_invoke_threads(*args, **kwargs)
        else:
            return self.reverse_recurse_invoke_single(*args, **kwargs)

    def __str__(self):
        ret = ""
        for d in sorted(self.depset):
            t = self.core.get(d)
            s = "{}: {}\n".format(d, sorted(t.deps))
            ret += s
        ret = ret[:-1]
        return ret


class Target:
    __actions__ = {"actlist", "print"}

    def __init__(self, tgt, deps, weakdeps=[], actions=None, __help__=None, **kwargs):
        self.tgt = tgt
        self.deps = set(deps)
        self.weakdeps = set(weakdeps)
        for k, v in kwargs.items():
            setattr(self, k, v)

        if actions is not None:
            self.__actions__ = self.__actions__.union(set(actions))

        self.__help__ = __help__

    def get_deplist(self):
        return [self.core.get(d) for d in self.deps]

    def actlist(self):
        print(licant.util.get_actions(self))

    def print(self):
        print(self.__dict__)

    def hasaction(self, act):
        return act in self.__actions__

    def invoke(self, funcname, *args, critical=False, **kwargs):
        """Invoke func function or method, or mthod with func name for this target

                Поддерживается несколько разных типов func.
                В качестве func может быть вызвана внешняя функция с параметром текущей цели,
                или название локального метода.
                critical -- Действует для строкового вызова. Если данный attr отсутствует у цели,
                то в зависимости от данного параметра может быть возвращен None или выброшено исключение.
                """
        if core.runtime["trace"]:
            print(
                "TRACE: Invoke: tgt:{}, act:{}, args:{}, kwargs:{}".format(
                    self.tgt, funcname, args, kwargs
                )
            )

        # if funcname not in self.__actions__:
        # 	licant.error("Isn't action {}".format(licant.util.yellow(funcname)))

        func = getattr(self, funcname, None)
        if func is None:
            if critical:
                raise WrongAction(self, funcname)
            return None

        if isinstance(func, types.MethodType):
            # return licant.util.cutinvoke(func, *args, **kwargs)
            return func(*args, **kwargs)

        else:
            # return licant.util.cutinvoke(func, self, *args, **kwargs)
            return func(self, *args, **kwargs)

    def __repr__(self):
        """По умолчанию вывод Target на печать возвращает идентификатор цели"""
        return self.tgt


class UpdateStatus(Enum):
    Waiting = 0
    Keeped = 1
    Updated = 2


class UpdatableTarget(Target):
    __actions__ = Target.__actions__.union(
        {"recurse_update", "update", "update_if_need"}
    )

    def __init__(
        self,
        tgt,
        deps,
        default_action="recurse_update",
        update_status=UpdateStatus.Waiting,
        **kwargs
    ):
        Target.__init__(self, tgt, deps,
                        default_action=default_action, **kwargs)
        self.update_status = update_status

    def update(self):
        licant.error("Unoverrided update method")

    def self_need(self):
        return False

    def has_updated_depends(self):
        depends = self.get_deplist()

        for d in depends:
            if (
                not isinstance(d, UpdatableTarget)
                or d.update_status == UpdateStatus.Updated
            ):
                return True

            if d.update_status == UpdateStatus.Waiting:
                print(d)
                licant.error("Unwalked depends in UpdatableTarget")

        return False

    def update_if_need(self):
        if self.has_updated_depends() or self.self_need():  # self.invoke("self_need"):
            self.update_status = UpdateStatus.Updated
            return self.update()  # self.invoke("update")
        else:
            self.update_status = UpdateStatus.Keeped
            return True

    def recurse_update(self):
        stree = self.core.subtree(self.tgt)
        stree.reverse_recurse_invoke(
            ops="update_if_need", threads=core.runtime["threads"]
        )


class Routine(UpdatableTarget):
    __actions__ = {"recurse_update", "update", "actlist"}

    def __init__(self, func, deps=[], tgt=None, **kwargs):
        if tgt is None:
            tgt = func.__name__
        UpdatableTarget.__init__(self, tgt=tgt, deps=deps, **kwargs)
        self.func = func
        self.args = []

    def update(self):
        return self.func(*self.args)

    def self_need(self):
        return True

    def recurse_update(self, *args):
        self.args = args
        super().recurse_update()


def routine_decorator(func=None, deps=None):
    if inspect.isfunction(func):
        core.add(Routine(func))
        return func

    else:

        def decorator(func):
            core.add(Routine(func, deps=deps))
            return func

        return decorator


def print_targets_list(target, *args):
    if core.runtime["debug"]:
        print("print_targets_list args:{}".format(args))

    if len(core.targets) == 0:
        print("targets doesn't founded")
        return

    keys = sorted(core.targets.keys())

    if len(args) > 0:
        keys = [m for m in keys if re.search(args[0], m)]

    for k in keys:
        print(k)


def print_target_info(taget, *args):
    if len(args) == 0:
        licant.error("Need target mnemo")

    print("name:", core.get(args[0]))
    print("deps:", sorted(core.get(args[0]).deps))


def print_deps(taget, *args):
    if len(args) == 0:
        name = licant.cli.default_target
    else:
        name = args[0]

    lst = sorted(core.depends_as_set(name))
    for l in lst:
        print(l)


def print_subtree(target, tgt):
    print(core.subtree(tgt))


corediag_target = Target(
    tgt="corediag",
    deps=[],
    targets=print_targets_list,
    tgtinfo=print_target_info,
    subtree=print_subtree,
    printdeps=print_deps,
    actions={"targets", "tgtinfo", "subtree", "printdeps"},
    __help__="Core state info",
)

core.add(corediag_target)


def do(lst):
    core.get(lst[0]).invoke(lst[1], *lst[2:])


def get_target(name):
    return core.get(name)
